/**
* Copyright (c) 2026 Huawei Technologies Co., Ltd.
* This program is free software, you can redistribute it and/or modify it under the terms and conditions of
* CANN Open Software License Agreement Version 2.0 (the "License").
* Please refer to the License for details. You may not use this file except in compliance with the License.
* THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND, EITHER EXPRESS OR IMPLIED,
* INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT, MERCHANTABILITY, OR FITNESS FOR A PARTICULAR PURPOSE.
* See LICENSE in the root of the software repository for the full text of the License.
*/
/* !
* \file math.asc
* \brief
*/
#include <algorithm>
#include <cmath>
#include <iostream>
#include <iterator>
#include <vector>
#include "acl/acl.h"
#include "tiling/platform/platform_ascendc.h"
#include "simt_api/math_functions.h"
__global__ void math_custom(const float* x, float* z, uint64_t total_length)
{
// Calculate global thread ID
int32_t idx = blockIdx.x * blockDim.x + threadIdx.x;
// Maps to the row index of output tensor
if (idx >= total_length) {
return;
}
const float results[] = {tanf(x[idx]), tanhf(x[idx]), atanf(x[idx]), atanhf(x[idx]), expf(x[idx]),
logf(x[idx]), cosf(x[idx]), coshf(x[idx]), sinf(x[idx]), sinhf(x[idx])};
z[idx] = results[idx];
}
std::vector<float> run_math_functions(const std::vector<float>& x)
{
size_t total_byte_size = x.size() * sizeof(float);
int32_t device_id = 0;
aclrtStream stream = nullptr;
const uint8_t* x_host = reinterpret_cast<const uint8_t*>(x.data());
uint8_t* z_host = nullptr;
float* x_device = nullptr;
float* z_device = nullptr;
// Init
aclInit(nullptr);
aclrtSetDevice(device_id);
aclrtCreateStream(&stream);
// Malloc memory in host and device
aclrtMallocHost((void**)(&z_host), total_byte_size);
aclrtMalloc((void**)&x_device, total_byte_size, ACL_MEM_MALLOC_HUGE_FIRST);
aclrtMalloc((void**)&z_device, total_byte_size, ACL_MEM_MALLOC_HUGE_FIRST);
aclrtMemcpy(x_device, total_byte_size, x_host, total_byte_size, ACL_MEMCPY_HOST_TO_DEVICE);
// Calculate split parameters
uint32_t block_num = 1;
uint32_t thread_num_per_block = x.size();
uint32_t dyn_ubuf_size = 0; // No need to alloc dynamic memory.
// Call kernel function with <<<...>>>
math_custom<<<block_num, thread_num_per_block, dyn_ubuf_size, stream>>>(x_device, z_device, x.size());
aclrtSynchronizeStream(stream);
// Copy result from device to host
aclrtMemcpy(z_host, total_byte_size, z_device, total_byte_size, ACL_MEMCPY_DEVICE_TO_HOST);
std::vector<float> output((float*)z_host, (float*)(z_host + total_byte_size));
// Free memory
aclrtFree(x_device);
aclrtFree(z_device);
aclrtFreeHost(z_host);
// Deinitialize
aclrtDestroyStream(stream);
aclrtResetDevice(device_id);
aclFinalize();
return output;
}
uint32_t verify_result(const std::vector<float>& output, const std::vector<float>& golden)
{
auto print_tensor = [](const std::vector<float>& tensor, const char* name) {
constexpr size_t max_print_size = 20;
std::cout << name << ": ";
std::copy(
tensor.begin(), tensor.begin() + std::min(tensor.size(), max_print_size),
std::ostream_iterator<float>(std::cout, " "));
if (tensor.size() > max_print_size) {
std::cout << "...";
}
std::cout << std::endl;
};
print_tensor(output, "Output");
print_tensor(golden, "Golden");
constexpr float relative_tolerance = 1e-4f;
constexpr float absolute_tolerance = 1e-5f;
bool is_match =
output.size() == golden.size() &&
std::equal(output.begin(), output.end(), golden.begin(), [](float actual, float expected) {
return std::fabs(actual - expected) <= absolute_tolerance + relative_tolerance * std::fabs(expected);
});
if (is_match) {
std::cout << "[Success] Case accuracy is verification passed." << std::endl;
return 0;
} else {
std::cout << "[Failed] Case accuracy is verification failed!" << std::endl;
return 1;
}
}
int32_t main(int32_t argc, char* argv[])
{
std::vector<float> x = {0.1f, 0.2f, 0.3f, 0.4f, 0.5f, 0.6f, 0.7f, 0.8f, 0.9f, 1.0f};
std::vector<float> golden = {std::tan(x[0]), std::tanh(x[1]), std::atan(x[2]), std::atanh(x[3]), std::exp(x[4]),
std::log(x[5]), std::cos(x[6]), std::cosh(x[7]), std::sin(x[8]), std::sinh(x[9])};
std::vector<float> output = run_math_functions(x);
return verify_result(output, golden);
}